"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.react2angular = void 0;
const ngcomponent_1 = require("ngcomponent");
const React = require("react");
const client_1 = require("react-dom/client");
/**
 * Wraps a React component in Angular. Returns a new Angular component.
 *
 * Usage:
 *
 *   ```ts
 *   type Props = { foo: number }
 *   class ReactComponent extends React.Component<Props, S> {}
 *   const AngularComponent = react2angular(ReactComponent, ['foo'])
 *   ```
 */
function react2angular(Class, bindingNames = null, injectNames = []) {
    const names = bindingNames
        || (Class.propTypes && Object.keys(Class.propTypes))
        || [];
    return {
        bindings: names.reduce((a, v) => (Object.assign(Object.assign({}, a), { [v]: '<' })), {}),
        controller: ['$element', ...injectNames, class extends ngcomponent_1.default {
                static get $$ngIsClass() {
                    return true;
                }
                constructor($element, ...injectedProps) {
                    super();
                    this.$element = $element;
                    this.isDestroyed = false;
                    this.injectedProps = {};
                    injectNames.forEach((name, i) => {
                        this.injectedProps[name] = injectedProps[i];
                    });
                    this.root = (0, client_1.createRoot)($element[0]);
                }
                render() {
                    if (!this.isDestroyed) {
                        this.root.render(React.createElement(Class, Object.assign({}, this.props, this.injectedProps)));
                    }
                }
                componentWillUnmount() {
                    this.isDestroyed = true;
                    if (this.$element[0] && this.root) {
                        this.root.unmount();
                    }
                }
            }]
    };
}
exports.react2angular = react2angular;
//# sourceMappingURL=index.js.map